/*******************************************************************************
 * Copyright 2003 Seapine Software, Inc. All rights reserved.
 * 
 * IMPORTANT:  This Software is supplied to you by Seapine Software, Inc. (Seapine)
 * in consideration of your agreement to the following terms, and your use, installation,
 * modification or redistribution of this Software constitutes acceptance of these terms.
 * If you do not agree with these terms, please do not use, install, modify or redistribute
 * this Software.
 * 
 * In consideration of your agreement to abide by the following terms, and subject to these
 * terms, Seapine grants you a personal, non-exclusive license, under Seapine copyrights in
 * this original Seapine software (the Software), to use, reproduce, modify and redistribute
 * the Software, with or without modifications, in source and/or binary forms; provided that
 * if you redistribute the Software in its entirety and without modifications, you must retain
 * this notice and the following text and disclaimers in all such redistributions of the Software.
 * Neither the name, trademarks, service marks or logos of Seapine, Inc. may be used to endorse
 * or promote products derived from the Software without specific prior written permission from
 * Seapine.  Except as expressly stated in this notice, no other rights or licenses, express or
 * implied, are granted by Seapine herein, including but not limited to any patent rights that
 * may be infringed by your derivative works or by other works in which the Software may be
 * incorporated.
 * 
 * The Software is provided by Seapine on an "AS IS" basis.  SEAPINE MAKES NO WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF NON-INFRINGEMENT,
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE SOFTWARE OR ITS USE AND
 * OPERATION ALONE OR IN COMBINATION WITH YOUR PRODUCTS. 
 * 
 * IN NO EVENT SHALL SEAPINE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE,
 * REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION OF THE SOFTWARE, HOWEVER CAUSED AND WHETHER
 * UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN
 * IF SEAPINE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>   // for EEXIST
#include <fcntl.h>   // following 4 are for _open() file locking
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>

#include <windows.h> // for Sleep()

//////////////////////////////////////////////////////////////////////////////

/*****
 Before compiling this trigger script, you will want to configure each of the
 variables as described below.  Many defaults may be acceptable.
   sSSCMExecutable - The full path to the sscm.exe executable
   sFTPExecutable - The full path to the Windows ftp.exe executable
   sErrorLog - A full path to an error log (or NULL for no error log).  This
               log holds all of the errors during the ftp process.
   sActivityLog - A full path to an activity log (or NULL for no activity
                  log).  This log holds all of the output of the sscm.exe
                  and ftp.exe command line activity.
   sLockFile - A full path to a lock file used to keep the command line
               clients from blocking on the activity log.
   sFTPCommandFile - A file to use to temporarily hold all of the FTP
                     commands to send to the FTP server.
   sTempDir - A location that can be used to store temporary files.
   sSCMServer - The IP:port of the SCM server to get files from.
   sSCMUser - The username to login to the SCM server as.
   sSCMUserPassword - The password for the above username.
   sSCMRepository - The topmost repository that will be sent to the FTP
                    server.
   sFTPServer - The IP address of the FTP server to upload to.
   sFTPUser - The username to use to login to the FTP server.
   sFTPUserPassword - The password for the above FTP username.
   sFTPBasePath - The topmost path on the FTP server where files should be
                  stored.
*****/

const char sSSCMExecutable[] = "\"C:\\Program Files\\Seapine\\Surround SCM\\sscm.exe\"";
const char sFTPExecutable[] = "\"C:\\Windows\\System32\\ftp.exe\"";
const char *sErrorLog = "C:\\ftp\\FTPErrors.log";
const char *sActivityLog = "C:\\ftp\\FTPActivity.log";
const char sLockFile[] = "C:\\ftp\\FTPLock.lck";
const char sFTPCommandFile[] = "C:\\ftp\\FTPCommand.txt";
const char sTempDir[] = "C:\\ftp";

const char sSCMServer[] = "192.168.1.100:4900";
const char sSCMUser[] = "ftp";
const char sSCMUserPassword[] = "ftp";
const char sSCMRepository[] = "mainline";

const char sFTPServer[] = "192.168.1.101";
const char sFTPUser[] = "ftp";
const char sFTPUserPassword[] = "ftp";
const char sFTPBasePath[] = "/home/ftp";


/*****
 Before using this script, you will want to do the following:
 1. Compile this script.  This code was built with Visual C++ 6.0, but could
    be built by most standard Windows C++ compilers.  The command line used
    to compile this script was "cl.exe FTPTransfer.cpp".
 2. Configure the SCM server to have an "ftp" user, or similar.  Make sure the
    user name and password match the ones that were entered in the script.
    Also be sure that the user has access to get files out of the specified
    SCM repositories.
 3. Be sure that the username and password specified above to log into the FTP
    site is valid and has access to create directories and files within the
    FTP base path listed above.
 4. Be sure that the local directories exist that the log files will reside in
    and that the temporary directory exists.
 5. Setup a trigger to fire this script from Surround SCM.  The trigger should
    be set to only run in a specified branch, at the above listed SCM
    repository, on all "*" files.  It should fire "after" "(all)" events.  Be
    sure to put the full path to the FTPTransfer script in the "Run this
    script" field.
*****/

//////////////////////////////////////////////////////////////////////////////

// These variables are the ones retrieved from the environment that the
// SCM trigger sets for us.
const char *sTriggerDesc = NULL;
const char *sTriggerEvent = NULL;
const char *sTriggerBeforeAfter = NULL;
const char *sTriggerDate = NULL;
const char *sTriggerTime = NULL;
const char *sTriggerUser = NULL;
const char *sTriggerClientIP = NULL;
const char *sTriggerClientMachine = NULL;
const char *sTriggerMainline = NULL;
const char *sTriggerBranch = NULL;
const char *sTriggerBranchType = NULL;
const char *sTriggerRepository = NULL;
const char *sTriggerFile = NULL;
const char *sTriggerFileVersion = NULL;
const char *sTriggerLocalFile = NULL;
const char *sTriggerComment = NULL;
const char *sTriggerSourceBranch = NULL;
const char *sTriggerDefectNumber = NULL;
const char *sTriggerShareRepository = NULL;
const char *sTriggerOldFileName = NULL;
const char *sTriggerPermRemove = NULL;

//////////////////////////////////////////////////////////////////////////////

// These functions handle most of the internals of FTP commands, check the
// function header and internal comments for more details.

int DoFTPTransferFile(FILE *fpErrorLog);
int DoFTPRemoveFile(FILE *fpErrorLog);
int DoFTPRenameFile(FILE *fpErrorLog);

bool GetLockForCLI();
void ClearLockForCLI();
int fhLockFile = -1; // Handle to the lock file, the file is deleted when this is closed.

//////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
   int nReturn = 0;
   FILE *fpErrorLog = NULL;
   if(sErrorLog)
   {
      if(!(fpErrorLog = fopen(sErrorLog, "a")))
         return -1;  // Error opening error log, and no way to report it, quit.
   }

   // Get all of Surround SCM's environment variables (some may go unused)
   sTriggerDesc = getenv("SSCM_TRIGGERDESC");
   sTriggerEvent = getenv("SSCM_EVENT");
   sTriggerBeforeAfter = getenv("SSCM_BEFOREAFTER");
   sTriggerDate = getenv("SSCM_DATE");
   sTriggerTime = getenv("SSCM_TIME");
   sTriggerUser = getenv("SSCM_USER");
   sTriggerClientIP = getenv("SSCM_CLIENTIP");
   sTriggerClientMachine = getenv("SSCM_CLIENTMACHINE");
   sTriggerMainline = getenv("SSCM_MAINLINE");
   sTriggerBranch = getenv("SSCM_BRANCH");
   sTriggerBranchType = getenv("SSCM_BRANCHTYPE");
   sTriggerRepository = getenv("SSCM_REPOSITORY");
   sTriggerFile = getenv("SSCM_FILE");
   sTriggerFileVersion = getenv("SSCM_FILEVERSION");
   sTriggerLocalFile = getenv("SSCM_LOCALFILE");
   sTriggerComment = getenv("SSCM_COMMENT");
   sTriggerSourceBranch = getenv("SSCM_SOURCEBRANCH");
   sTriggerDefectNumber = getenv("SSCM_DEFECTNUMBER");
   sTriggerShareRepository = getenv("SSCM_SHAREREPOSITORY");
   sTriggerOldFileName = getenv("SSCM_OLDFILENAME");
   sTriggerPermRemove = getenv("SSCM_PERMREMOVE");

   // Check to make sure we have the env vars that we need
   if(!sTriggerDate || !sTriggerTime)      
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "!! !! : One or more trigger environment variables was invalid.\n");
      nReturn = -1;
   }
   else if(!sTriggerEvent || !sTriggerUser || !sTriggerClientIP ||
           !sTriggerClientMachine || !sTriggerBranch ||
           !sTriggerRepository || !sTriggerFile)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : One or more trigger environment variables was invalid.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
   }

   // Start a transfer for the appropriate event
   if(nReturn == 0)
   {
      if(strncmp(sSCMRepository, sTriggerRepository, strlen(sSCMRepository)) != 0)
         nReturn = 0; // Not in our repository
      else if(strcmp(sTriggerEvent, "Add") == 0)
         nReturn = DoFTPTransferFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Add from branch") == 0)
         nReturn = DoFTPTransferFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Attach to defect") == 0)
         nReturn = 0; // Nothing to do
      else if(strcmp(sTriggerEvent, "Check in") == 0)
         nReturn = DoFTPTransferFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Promote from") == 0)
         nReturn = DoFTPTransferFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Rebase") == 0)
         nReturn = DoFTPTransferFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Remove") == 0)
         nReturn = DoFTPRemoveFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Rename file") == 0)
         nReturn = DoFTPRenameFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Restore") == 0)
         nReturn = DoFTPTransferFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Share broken") == 0)
         nReturn = 0; // Nothing to do
      else if(strcmp(sTriggerEvent, "Share created") == 0)
         nReturn = DoFTPTransferFile(fpErrorLog);
      else
      {
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Trigger event does not match any known event type.\n",
                                sTriggerDate, sTriggerTime);
         nReturn = -1;  // Some unknown event, we should never hit this
      }
   }

   if(fpErrorLog)
      fclose(fpErrorLog);
   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles sending a file to the FTP server
int DoFTPTransferFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Try and lock files so we can do our transfer uninterrupted
   if(!GetLockForCLI())
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error obtaining lock file.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
      return nReturn;
   }

   // Open our FTP command file
   FILE *fpFTPCommandFile = NULL;
   if(!(fpFTPCommandFile = fopen(sFTPCommandFile, "wt")))
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Unable to open FTP command file.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
      ClearLockForCLI();
      return nReturn;
   }

   // Setup the FTP login
   fprintf(fpFTPCommandFile, "open %s\n", sFTPServer);
   fprintf(fpFTPCommandFile, "%s\n", sFTPUser);
   fprintf(fpFTPCommandFile, "%s\n", sFTPUserPassword);

   // Make (recursively) and change into the directory needed to put the file
   fprintf(fpFTPCommandFile, "cd %s\n", sFTPBasePath);
   if(strlen(sTriggerRepository) != strlen(sSCMRepository)) // not base dir?
   {
      char *sMkdir = new char[strlen(sTriggerRepository)+1];
      sMkdir[0] = 0;
      // Copy the repository in, skipping over the parent part
      strcpy(sMkdir, sTriggerRepository + strlen(sSCMRepository) + 1);

      char *a = sMkdir;
      char *b = sMkdir;
      bool bDone = false;
      while(!bDone)
      {
         while(*a && *a != '/')
            a++;
         if(!*a)
            bDone = true;
         *a = 0;
         fprintf(fpFTPCommandFile, "mkdir %s\n", b);
         fprintf(fpFTPCommandFile, "cd %s\n", b);
         b = a + 1;
         a++;
      }
      delete [] sMkdir;
   }

   // Get the file off of our SCM server
   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " get ");
   strcat(sCmdLine, sTriggerFile);
   strcat(sCmdLine, " -b");
   strcat(sCmdLine, sTriggerBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   strcat(sCmdLine, " -d");
   strcat(sCmdLine, sTempDir);
   strcat(sCmdLine, " -f -e -wreplace -y");
   strcat(sCmdLine, sSCMUser);
   strcat(sCmdLine, ":");
   strcat(sCmdLine, sSCMUserPassword);
   strcat(sCmdLine, " -z");
   strcat(sCmdLine, sSCMServer);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
   }
   nReturn = system(sCmdLine);
   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing get operation on %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      fclose(fpFTPCommandFile);
   }
   else
   {
      // Get was good, FTP the file to the remote server
      fprintf(fpFTPCommandFile, "put \"%s\\%s\"\n", sTempDir, sTriggerFile);
      fprintf(fpFTPCommandFile, "quit\n");
      fclose(fpFTPCommandFile);
      strcpy(sCmdLine, sFTPExecutable);
      strcat(sCmdLine, " -s:");
      strcat(sCmdLine, sFTPCommandFile);
      if(sActivityLog)
      {
         strcat(sCmdLine, " >> ");
         strcat(sCmdLine, sActivityLog);
      }
      nReturn = system(sCmdLine);
      if(nReturn)
      {
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error performing FTP transfer of %s/%s.\n",
                                sTriggerDate, sTriggerTime,
                                sTriggerRepository, sTriggerFile);
      }
   }

   // Get rid of the command file and our local file
   remove(sFTPCommandFile);
   strcpy(sCmdLine, sTempDir);
   strcat(sCmdLine, "\\");
   strcat(sCmdLine, sTriggerFile);
   remove(sCmdLine);

   // Clear our lock
   ClearLockForCLI();

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles removal of a remote FTP file
int DoFTPRemoveFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Try and lock files so we can do our transfer uninterrupted
   if(!GetLockForCLI())
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error obtaining lock file.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
      return nReturn;
   }

   // Open our FTP command file
   FILE *fpFTPCommandFile = NULL;
   if(!(fpFTPCommandFile = fopen(sFTPCommandFile, "wt")))
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Unable to open FTP command file.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
      ClearLockForCLI();
      return nReturn;
   }

   // Setup the FTP login
   fprintf(fpFTPCommandFile, "open %s\n", sFTPServer);
   fprintf(fpFTPCommandFile, "%s\n", sFTPUser);
   fprintf(fpFTPCommandFile, "%s\n", sFTPUserPassword);

   // Change to the FTP directory where the file is
   if(strlen(sTriggerRepository) == strlen(sSCMRepository))
      fprintf(fpFTPCommandFile, "cd %s\n", sFTPBasePath);
   else
      fprintf(fpFTPCommandFile, "cd %s/%s\n", sFTPBasePath, sTriggerRepository + strlen(sSCMRepository) + 1);

   // Remove the file
   fprintf(fpFTPCommandFile, "delete \"%s\"\n", sTriggerFile);
   fprintf(fpFTPCommandFile, "quit\n");
   fclose(fpFTPCommandFile);
   strcpy(sCmdLine, sFTPExecutable);
   strcat(sCmdLine, " -s:");
   strcat(sCmdLine, sFTPCommandFile);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
   }
   nReturn = system(sCmdLine);
   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing FTP remove of %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
   }

   // Get rid of the command file
   remove(sFTPCommandFile);

   // Clear our lock
   ClearLockForCLI();

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles renaming of a remote FTP file
int DoFTPRenameFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Try and lock files so we can do our transfer uninterrupted
   if(!GetLockForCLI())
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error obtaining lock file.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
      return nReturn;
   }

   // Open our FTP command file
   FILE *fpFTPCommandFile = NULL;
   if(!(fpFTPCommandFile = fopen(sFTPCommandFile, "wt")))
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Unable to open FTP command file.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
      ClearLockForCLI();
      return nReturn;
   }

   // Setup the FTP login
   fprintf(fpFTPCommandFile, "open %s\n", sFTPServer);
   fprintf(fpFTPCommandFile, "%s\n", sFTPUser);
   fprintf(fpFTPCommandFile, "%s\n", sFTPUserPassword);

   // Change to the FTP directory where the file is
   if(strlen(sTriggerRepository) == strlen(sSCMRepository))
      fprintf(fpFTPCommandFile, "cd %s\n", sFTPBasePath);
   else
      fprintf(fpFTPCommandFile, "cd %s/%s\n", sFTPBasePath, sTriggerRepository + strlen(sSCMRepository) + 1);

   // Rename the file
   fprintf(fpFTPCommandFile, "rename \"%s\" \"%s\"\n", sTriggerOldFileName, sTriggerFile);
   fprintf(fpFTPCommandFile, "quit\n");
   fclose(fpFTPCommandFile);
   strcpy(sCmdLine, sFTPExecutable);
   strcat(sCmdLine, " -s:");
   strcat(sCmdLine, sFTPCommandFile);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
   }
   nReturn = system(sCmdLine);
   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing FTP rename of %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
   }

   // Get rid of the command file
   remove(sFTPCommandFile);

   // Clear our lock
   ClearLockForCLI();

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// This function will wait up to 30 seconds to get a lock on the lock file.
// It returns false if it could not get a lock.
bool GetLockForCLI()
{
   bool bReturn = false;
   int nSecsToKeepTrying = 30000;  // Wait up to 30.000 seconds for a lock
   if(sLockFile)
   {
      while(nSecsToKeepTrying > 0)
      {
         fhLockFile = _open(sLockFile, _O_CREAT | _O_EXCL | _O_TEMPORARY, _S_IWRITE);
         if(fhLockFile == -1  &&  errno == EEXIST)
         {
            // Lock file is in use by someone else... wait for it.
            nSecsToKeepTrying -= 100;
            Sleep(100);
         }
         else if(fhLockFile != -1)
         {
            // Lock file is good
            bReturn = true;
            nSecsToKeepTrying = 0;
         }
         else
         {
            // Some error with the lock file, just quit
            bReturn = false;
            nSecsToKeepTrying = 0;
         }
      }
   }

   return bReturn;
}

//////////////////////////////////////////////////////////////////////////////

// This function clears the lock by closing the file handle (which deletes
// the file).
void ClearLockForCLI()
{
   if(fhLockFile != -1)
      _close(fhLockFile);
}

//////////////////////////////////////////////////////////////////////////////
